@@ -3,7 +3,7 @@
    using STATE as the random state previously initialized by a call to
    gmp_randinit().
 
-Copyright (C) 1999, 2000  Free Software Foundation, Inc.
+Copyright 1999, 2000, 2001, 2002  Free Software Foundation, Inc.
 
 This file is part of the GNU MP Library.
 
@@ -26,30 +26,22 @@ MA 02111-1307, USA. */
 #include "gmp-impl.h"
 
 void
-#if __STDC__
 mpf_urandomb (mpf_t rop, gmp_randstate_t rstate, unsigned long int nbits)
-#else
-mpf_urandomb (rop, rstate, nbits)
-     mpf_t rop;
-     gmp_randstate_t rstate;
-     unsigned long int nbits;
-#endif
 {
   mp_ptr rp;
   mp_size_t nlimbs;
   mp_exp_t exp;
 
   rp = PTR (rop);
-  nlimbs = (nbits + BITS_PER_MP_LIMB - 1) / BITS_PER_MP_LIMB;
+  nlimbs = (nbits + GMP_NUMB_BITS - 1) / GMP_NUMB_BITS;
 
   _gmp_rand (rp, rstate, nbits);
 
-  /* If nbits isn't a multiple of BITS_PER_MP_LIMB, shift up.  */
+  /* If nbits isn't a multiple of GMP_NUMB_BITS, shift up.  */
   if (nlimbs != 0)
     {
-      if (nbits % BITS_PER_MP_LIMB != 0)
-	mpn_lshift (rp, rp, nlimbs,
-		    BITS_PER_MP_LIMB - nbits % BITS_PER_MP_LIMB);
+      if (nbits % GMP_NUMB_BITS != 0)
+	mpn_lshift (rp, rp, nlimbs, GMP_NUMB_BITS - nbits % GMP_NUMB_BITS);
     }
 
   exp = 0;
@@ -60,5 +52,4 @@ mpf_urandomb (rop, rstate, nbits)
     }
   EXP (rop) = exp;
   SIZ (rop) = nlimbs;
-
 }
